;;; -*- Mode: LISP; Syntax: ANSI-Common-Lisp; Package: (C-INTERFACE); -*-

;;; C-Interface
;;; Copyright (C) 2002 Matthias Hlzl

;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License as
;;; published by the Free Software Foundation; either version 2 of the
;;; License, or (at your option) any later version.

;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.

;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
;;; 02111-1307, USA.


(in-package "C-INTERFACE")

(declaim (optimize (debug 3)))

;;; C-Type-References.
;;; =================

(defclass named-reference-mixin ()
  ((lisp-name :accessor lisp-name :initarg :lisp-name)))

(defclass c-type-reference ()
  ())

(defclass primitive-c-type-reference (c-type-reference)
  ())

(defclass complex-c-type-reference (c-type-reference)
  ())

(defclass named-c-type-reference (complex-c-type-reference named-reference-mixin)
  ())

(defclass c-struct-type-reference (named-c-type-reference)
  ())

(defclass c-union-type-reference (named-c-type-reference)
  ())

(defclass c-enum-type-reference (named-c-type-reference)
  ())

(defclass c-function-type-reference (complex-c-type-reference)
  ((arglist :accessor arglist :initarg :arglist :initform '())
   (result-type :accessor result-type :initarg :result-type)))

(defclass derived-c-type-reference (complex-c-type-reference)
  ((base-type-reference :accessor base-type-reference
			:initarg :base-type-reference)))

(defclass c-pointer-type-reference (derived-c-type-reference)
  ())

(defclass c-array-type-reference (derived-c-type-reference)
  ((array-size :accessor array-size :initarg :array-size)))

(defclass c-typedef-reference (named-c-type-reference)
  ())


(defclass c-char (primitive-c-type-reference)
  ())

(defclass signed-c-char (primitive-c-type-reference)
  ())

(defclass unsigned-c-char (primitive-c-type-reference)
  ())

(defclass c-short (primitive-c-type-reference)
  ())

(defclass signed-c-short (primitive-c-type-reference)
  ())

(defclass unsigned-c-short (primitive-c-type-reference)
  ())

(defclass c-int (primitive-c-type-reference)
  ())

(defclass signed-c-int (primitive-c-type-reference)
  ())

(defclass unsigned-c-int (primitive-c-type-reference)
  ())

(defclass c-long (primitive-c-type-reference)
  ())

(defclass signed-c-long (primitive-c-type-reference)
  ())

(defclass unsigned-c-long (primitive-c-type-reference)
  ())

(defclass c-long-long (primitive-c-type-reference)
  ())

(defclass signed-c-long-long (primitive-c-type-reference)
  ())

(defclass unsigned-c-long-long (primitive-c-type-reference)
  ())

(defclass c-float (primitive-c-type-reference)
  ())

(defclass c-double (primitive-c-type-reference)
  ())

(defclass c-void (primitive-c-type-reference)
  ())

;;; Alien-Type-For.
;;; ==============

(defgeneric alien-type-for (c-type-reference context))

(defmethod alien-type-for ((type c-struct-type-reference) context)
  (declare (ignorable context))
  (if (ignored-struct-p (lisp-name type))
      'c-call:unsigned-long
      `(alien:struct ,(lisp-name type))))

(defmethod alien-type-for ((type c-union-type-reference) context)
  (declare (ignorable context))
  `(alien:union ,(lisp-name type)))

(defmethod alien-type-for ((type c-enum-type-reference) context)
  (declare (ignorable context))
  `(alien:enum ,(lisp-name type)))

(defmethod alien-type-for ((type c-pointer-type-reference) context)
  (if (eq context :function-argument)
      '(alien:* t)
      `(alien:* ,(alien-type-for (base-type-reference type) context)))
  #+nil
  (declare (ignore context))
  #+nil
  'system:system-area-pointer)

(defmethod alien-type-for ((type c-array-type-reference) context)
  (let* ((size (array-size type))
	 (size-list (when (and size (not (zerop size)))
		      (list size))))
    `(alien:array
      ,(alien-type-for (base-type-reference type) context)
      ,@size-list)))

(defmethod alien-type-for ((type c-function-type-reference) context)
  (declare (ignorable context))
  `(alien:function
    ,(alien-type-for (result-type type) :function-return)
    ,@(mapcar (lambda (arg)
		(alien-type-for arg :function-argument))
	      (arglist type))))

(defmethod alien-type-for ((type c-char) context)
  (declare (ignorable context))
  'c-call:char)

(defmethod alien-type-for ((type signed-c-char) context)
  (declare (ignorable context))
  'c-call:char)

(defmethod alien-type-for ((type unsigned-c-char) context)
  (declare (ignorable context))
  'c-call:unsigned-char)

(defmethod alien-type-for ((type c-int) context)
  (declare (ignorable context))
  'c-call:int)

(defmethod alien-type-for ((type signed-c-int) context)
  (declare (ignorable context))
  'c-call:int)

(defmethod alien-type-for ((type unsigned-c-int) context)
  (declare (ignorable context))
  'c-call:unsigned-int)

(defmethod alien-type-for ((type c-short) context)
  (declare (ignorable context))
  'c-call:short)

(defmethod alien-type-for ((type signed-c-short) context)
  (declare (ignorable context))
  'c-call:short)

(defmethod alien-type-for ((type unsigned-c-short) context)
  (declare (ignorable context))
  'c-call:unsigned-short)

(defmethod alien-type-for ((type c-long) context)
  (declare (ignorable context))
  'c-call:long)

(defmethod alien-type-for ((type signed-c-long) context)
  (declare (ignorable context))
  'c-call:long)

(defmethod alien-type-for ((type unsigned-c-long) context)
  (declare (ignorable context))
  'c-call:unsigned-long)

(defmethod alien-type-for ((type c-long-long) context)
  (declare (ignorable context))
  (warn "Trying to convert unsupported long long C-type-reference to alien.")
  'c-call:long)

(defmethod alien-type-for ((type signed-c-long-long) context)
  (declare (ignorable context))
  (warn "Trying to convert unsupported long long C-type-reference to alien.")
  'c-call:long)

(defmethod alien-type-for ((type unsigned-c-long-long) context)
  (declare (ignorable context))
  (warn "Trying to convert unsupported long long C-type-reference to alien.")
  'c-call:unsigned-long)

(defmethod alien-type-for ((type c-float) context)
  (declare (ignorable context))
  'c-call:float)

(defmethod alien-type-for ((type c-double) context)
  (declare (ignorable context))
  'c-call:double)

(defmethod alien-type-for ((type c-void) context)
  (case context
    (:function-return 'c-call:void)
    ;;; ### This is probably not correct.  Need to figure out how to
    ;;; handle varargs.
    (t '(alien:* t)))) ; #+nil'system:system-area-pointer

;;; There are two possible choices: Output the name of the typedef and
;;; define an alien typedef as well, or translate the typedef to the
;;; base type in the conversion process.  I have chosen the former
;;; approach for the moment.
;;;
(defmethod alien-type-for ((type c-typedef-reference) context)
  (declare (ignorable context))
  (lisp-name type))

;;; Type-For.
;;; ========

;;; Generate a type that is suitable for a function declaration.

(defgeneric type-for (type))

(defmethod type-for (type)
  `(alien:alien ,(alien-type-for type nil)))

(defmethod type-for ((type c-pointer-type-reference))
  `(alien:alien ,(alien-type-for type nil)))

(defmethod type-for ((type c-typedef-reference))
  `(alien:alien ,(alien-type-for type nil)))

;;; Processing FFI Type Specifications.
;;; ==================================

(defgeneric process-typespec (c-type typespec))

(defmethod process-typespec (c-type typespec)
  (declare (ignorable typespec))
  c-type)

(defmethod process-typespec :before ((ref named-reference-mixin) typespec)
  (setf (lisp-name ref) (translate-c-name (second typespec))))

(defmethod process-typespec ((ref c-function-type-reference) typespec)
  (destructuring-bind (function arglist result-type) typespec
    (assert (eq function 'function))
    (if (eq (first arglist) 'void)
	(setf (arglist ref) '())
	(setf (arglist ref) (mapcar #'convert-ffi-type arglist)))
    (setf (result-type ref) (convert-ffi-type result-type))))

(defmethod process-typespec ((ref derived-c-type-reference) typespec)
  (destructuring-bind (ref-type base-type &rest stuff) typespec
    (declare (ignore ref-type stuff))
    (setf (base-type-reference ref) (convert-ffi-type base-type))))

(defmethod process-typespec ((ref c-array-type-reference) typespec)
  (destructuring-bind (type size base-type) typespec
    (assert (eq type 'array))
    (setf (base-type-reference ref) (convert-ffi-type base-type)
	  (array-size ref) size)))

;;; Conversion from FFIGen representation to C-Types.
;;; =================================================

(defparameter *ffi-type-to-c-type-translations*
  '(;;; Primitive Types.
    (char c-char) (signed-char signed-c-char) (unsigned-char unsigned-c-char)
    (short c-short) (signed-short signed-c-short) (unsigned-short unsigned-c-short)
    (int c-int) (signed-int signed-c-int) (unsigned-int unsigned-c-int)
    (unsigned unsigned-c-int) (signed signed-c-int)
    (long c-long) (signed-long signed-c-long) (unsigned-long unsigned-c-long)
    (long-long c-long-long) (signed-long-long signed-c-long-long)
    (unsigned-long-long unsigned-c-long-long)
    (float c-float) (double c-double)
    (void c-void)
    ;; Complex Types.
    (typedef c-typedef-reference)
    (struct-ref c-struct-type-reference)
    (union-ref c-union-type-reference)
    (enum-ref c-enum-type-reference)
    (function c-function-type-reference)
    (pointer c-pointer-type-reference)
    (array c-array-type-reference)))

(defun c-type-for (ffi-type)
  (second (find ffi-type *ffi-type-to-c-type-translations* :key #'first)))

(defun convert-ffi-type (typespec)
  (let ((c-type (c-type-for (first typespec))))
    (if c-type
	(let ((result (make-instance c-type)))
	  (process-typespec result typespec)
	  result)
	(error "Don't know how to convert type ~A." typespec))))