;;; -*- Mode: LISP; Syntax: ANSI-Common-Lisp; Package: (C-INTERFACE); -*-

;;; C-Interface
;;; Copyright (C) 2002 Matthias Hlzl

;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License as
;;; published by the Free Software Foundation; either version 2 of the
;;; License, or (at your option) any later version.

;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.

;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
;;; 02111-1307, USA.

(in-package "C-INTERFACE")

(declaim (optimize (debug 3)))

;;; The classes containing the first intermediate representation.

(define-hashed-class ffi-object ()
  ((name :initarg :name :accessor name)
   (lisp-name :initarg :lisp-name :accessor lisp-name)
   (data :initarg :data :accessor data)
   (source-location :initarg :source-location :accessor source-location)
   (visibility :initarg :visibility :accessor visibility)
   (processing-state :initform :unprocessed :accessor processing-state)))

(defclass base-type-mixin ()
  ((base-type :initarg :base-type :accessor base-type)))

(defclass field-item ()
  ((item-name :accessor item-name :initarg :item-name)
   (item-type :accessor item-type :initarg :item-type)
   (item-offset :accessor item-offset :initarg :item-offset)
   (item-byte-size :accessor item-byte-size :initarg :item-byte-size)))

(defclass field-mixin ()
  ((fields :initarg :fields :accessor fields)))

(defmethod fields (object)
  (declare (ignorable object))
  nil)

(define-hashed-class ffi-function (ffi-object)
  ((arglist :initarg :arglist :accessor arglist)
   (result-type :initarg :result-type :accessor result-type)
   (function-type :initarg :function-type :accessor function-type)))

(define-hashed-class ffi-variable (ffi-object base-type-mixin)
  ())

(define-hashed-class ffi-typedef (ffi-object base-type-mixin)
  ())

(defmethod fields ((typedef ffi-typedef))
  (let ((base-type (base-type typedef)))
    #+nil
    (when (typep base-type 'named-reference-mixin)
      (fields (find-ffi-object (lisp-name base-type))))
    (fields base-type)))

(define-hashed-class ffi-struct (ffi-object field-mixin)
  ())

(define-hashed-class ffi-union (ffi-object field-mixin)
  ())

(define-hashed-class ffi-macro (ffi-object)
  ((expansion :initarg :expansion :accessor expansion :initform nil)))

(defclass enum-item ()
  ((item-name :accessor item-name :initarg :item-name)
   (item-value :accessor item-value :initarg :item-value)))

(define-hashed-class ffi-enum (ffi-object)
  ((enum-items :accessor enum-items :initarg :enum-items :initform '())))

(define-hashed-class ffi-enum-identifier (ffi-object)
  ())

(defun type-for-form (form)
  (ccase (car form)
    (function 'ffi-function)
    (var 'ffi-variable)
    (type 'ffi-typedef)
    (struct 'ffi-struct)
    (union 'ffi-union)
    (macro 'ffi-macro)
    (enum 'ffi-enum)
    (enum-ident 'ffi-enum-identifier)))

(defun parse-ffi-form (form &key (type (type-for-form form)))
  (destructuring-bind (ignore source-location name expansion &optional visibility)
      form
    (declare (ignore ignore))
    (let ((lisp-name (translate-c-name name :type type)))
      (make-instance type
		     :source-location source-location
		     :name name
		     :lisp-name lisp-name
		     :data expansion
		     :visibility visibility
		     :id lisp-name))))

(defun parse-ffi-file (filename)
  (with-open-file (stream filename :direction :input)
    (loop for form = (read stream nil nil)
	  while form
	  do (parse-ffi-form form))))

(defun show-ffi-objects ()
  (let ((count 0)
        (states ()))
    (maphash (lambda (key values)
	       (mapc (lambda (value)
		       (let ((state (processing-state value)))
			 (pushnew state states)
			 (when (eq state :failed)
			   (format t "~&~A failed.~%" key)))
		       (incf count))
		     values))
             (get-ffi-object-table))
    (format t "~&~A objects, states: ~A.~%" count states)))

(defun clear-all-ffi-tables ()
  (clear-ffi-object-table)
  (clear-ffi-function-table)
  (clear-ffi-variable-table)
  (clear-ffi-typedef-table)
  (clear-ffi-struct-table)
  (clear-ffi-union-table)
  (clear-ffi-macro-table)
  (clear-ffi-enum-table)
  (clear-ffi-enum-identifier-table)
  (clear-c-name-translations))

;;; FFI-Object-For-Reference.
;;; ========================

(defgeneric ffi-object-for-reference (c-type-reference))

(defmethod ffi-object-for-reference ((ref primitive-c-type-reference))
  ;; There is no FFI-Object for primitive types...
  nil)

(defmethod ffi-object-for-reference ((ref c-struct-type-reference))
  (first (find-ffi-struct (lisp-name ref))))

(defmethod ffi-object-for-reference ((ref c-union-type-reference))
  (first (find-ffi-union (lisp-name ref))))

(defmethod ffi-object-for-reference ((ref c-enum-type-reference))
  (first (find-ffi-enum (lisp-name ref))))

(defmethod ffi-object-for-reference ((ref c-function-type-reference))
  (warn "Trying to find FFI-Object for function type.")
  nil)

(defmethod ffi-object-for-reference ((ref c-pointer-type-reference))
  (warn "Trying to find FFI Object for pointer type.")
  nil)

(defmethod ffi-object-for-reference ((ref c-array-type-reference))
  (warn "Trying to find FFI Object for array type.")
  nil)

(defmethod ffi-object-for-reference ((ref c-typedef-reference))
  (first (find-ffi-typedef (lisp-name ref))))

#||
(parse-ffi-file "/home/tc/prog/lisp/own-work/c-interface/ffi-files/test.ffi")
(parse-ffi-file "/home/tc/prog/lisp/own-work/c-interface/ffi-files/Xlib.ffi")
(parse-ffi-file "/home/tc/prog/lisp/own-work/c-interface/ffi-files/glx.ffi")
(parse-ffi-file "/home/tc/prog/lisp/own-work/c-interface/ffi-files/glu.ffi")
(show-ffi-objects)

(maphash (lambda (k vs)
           (mapcar (lambda (v)
		     (format t "~&~A ~A, state: ~S~%"
			     (type-of v) k (processing-state v)))
		   vs))
         (get-ffi-object-table))
||#
