;;; -*- Mode: LISP; Syntax: ANSI-Common-Lisp; Package: (C-INTERFACE); -*-

;;; C-Interface
;;; Copyright (C) 2002 Matthias Hlzl

;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License as
;;; published by the Free Software Foundation; either version 2 of the
;;; License, or (at your option) any later version.

;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.

;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
;;; 02111-1307, USA.

(in-package "C-INTERFACE")

;;; Alien-Form-For.
;;; ==============

;;; The generic function Alien-Form-for returns a Lisp form that is
;;; suitable as the CMUCL representation for a FFI-Object.

(defgeneric alien-form-for (ffi-object)
  (:documentation
   "Converts FFI-Object to a list-representation of an alien type."))

(defvar *conversion-impossible-actions*
  '((:not-implemented :warn)
    (:failed :error)
    (:in-progress :error))
  "An alist mapping ffi-object states for which no conversion is
possible to action descriptions.  Possible action descriptions are
  :ignore -- do nothing
  :warn   -- signal a warning and continue translation
  :error  -- signal an error.")

(defun action-for-state (processing-state)
  "Returns an action description for Processing-State.  When the
Processing-State is :Done, Nil is returned, otherwise the returned
value is determined by *Conversion-Impossible-Actions*."
  (if (eq processing-state :done)
      :ignore
      (second (assoc processing-state *conversion-impossible-actions*))))

(defun conversion-impossible (ffi-object)
  "Called when conversion of a FFI-Object that was not processed
correctly is attempted."
  (let* ((state (processing-state ffi-object))
	 (action (action-for-state state)))
    (ccase action
      (:ignore nil)
      (:warn (warn "Cannot convert ~A: ~A." ffi-object state))
      (:error (error "Cannot convert ~A: ~A." ffi-object state)))))

(defun generate-argument-vars (arglist)
  (loop for i from 0 below (length arglist)
	collect (intern (format nil "ARG-~A" i))))

(defun alien-form-for-fields (fields)
  (mapcar (lambda (field)
	    `(,(item-name field)
	      ,(alien-type-for (item-type field) :field-type)))
	  fields))

#+nil
(defun alien-form-for-field-accessor (prefix)
  (lambda (field-item)
    (let ((accessor-name (ext:symbolicate prefix (item-name field-item)))
	  (sap-accessor (sap-accessor-for-byte-size (item-byte-size field-item))))
      (when sap-accessor
	`((defun ,accessor-name (sap)
	    (declare (type system:system-area-pointer sap))
	    (,sap-accessor
	     sap ,(item-offset field-item)))
	  (defun (setf ,accessor-name) (new-value sap)
	    (declare (type system:system-area-pointer sap))
	    (setf (,sap-accessor sap ,(item-offset field-item))
		  new-value)))))))

(defun alien-form-for-field-accessor (prefix)
  (lambda (field-item)
    (let ((accessor-name (ext:symbolicate prefix (item-name field-item))))
      `((declaim (inline ,accessor-name (setf ,accessor-name)))
	(defun ,accessor-name (alien)
	  (alien:slot alien ',(item-name field-item)))
	(defun (setf ,accessor-name) (value alien)
	  (setf (alien:slot alien ',(item-name field-item))
		value))))))

(defun alien-forms-for-field-accessors (record)
  (let ((fields (fields record))
	(prefix (concatenate 'string
			     (symbol-name (lisp-name record))
			     "-")))
    (when fields
      (mapcan (alien-form-for-field-accessor prefix) fields))))


(defmethod alien-form-for ((function ffi-function))
  (let* ((argument-types (mapcar (lambda (arg)
				   (alien-type-for arg :function-argument))
				 (arglist function)))
	 (result-type (alien-type-for (result-type function) :function-return))
	 (lisp-name (lisp-name function))
	 (c-name (name function))
	 (argument-vars (generate-argument-vars argument-types)))
    `(progn
      (declaim (inline ,lisp-name))
      (defun ,lisp-name ,argument-vars
	(declare ,@(mapcar (lambda (type name)
			     `(type ,(type-for type) ,name))
			   (arglist function) argument-vars)
		 #+nil
		 (values ,(type-for (result-type function))))
	(values (alien:alien-funcall
		 (alien:extern-alien ,c-name (function ,result-type ,@argument-types))
		 ,@argument-vars))))))

(defmethod alien-form-for ((variable ffi-variable))
  `(alien:def-alien-variable (,(lisp-name variable) ,(name variable))
    ,(alien-type-for (base-type variable) nil)))

(defmethod alien-form-for ((typedef ffi-typedef))
  `(progn
     (alien:def-alien-type ,(lisp-name typedef)
	 ,(alien-type-for (base-type typedef) nil))
    ,@(alien-forms-for-field-accessors typedef)))

(defmethod alien-form-for ((struct ffi-struct))
  (if (ignored-struct-p struct)
      (format nil ";;; Ignored struct ~A.")
      `(progn
	(alien:def-alien-type nil
	    (alien:struct ,(lisp-name struct)
	     ,@(alien-form-for-fields (fields struct))))
	,@(alien-forms-for-field-accessors struct))))

(defmethod alien-form-for ((union ffi-union))
  `(progn
     (alien:def-alien-type nil ; ,(lisp-name union)
       (alien:union ,(lisp-name union)
	,@(alien-form-for-fields (fields union))))
    ,@(alien-forms-for-field-accessors union)))
    

(defmethod alien-form-for ((macro ffi-macro))
  (if (eq (processing-state macro) :done)
      `(defconstant ,(lisp-name macro) ,(expansion macro))
      (format nil ";;; Ignored macro ~A." (name macro))))

(defun alien-form-for-items (items)
  (mapcar (lambda (item)
	    (if (item-value item)
		`(,(item-name item) ,(item-value item))
		(item-name item)))
	  items))

(defmethod alien-form-for ((enum ffi-enum))
  `(alien:def-alien-type nil ; ,(lisp-name enum)
     (alien:enum ,(lisp-name enum)
      ,@(alien-form-for-items (enum-items enum)))))

(defmethod alien-form-for ((identifier ffi-enum-identifier))
  `(defconstant ,(lisp-name identifier) ,(data identifier)))

;;; Utilities.
;;; =========

(defun convert-form (form)
  "Converts Form to a list that describes Form's semantics as alien
structure."
  (alien-form-for (parse-and-process-ffi-form form)))


;;; Conversion of files.
;;; ===================

(defun all-exported-symbols ()
  (loop for name being each hash-key of (get-ffi-object-table)
	collect (symbol-name name)))

(defun print-defpackage (outstream package)
  (format outstream "(in-package \"COMMON-LISP-USER\")~2%")
  (format outstream "(defpackage ~W~%  (:use \"COMMON-LISP\")~%"
	  package)
  (format outstream "  ~< ~W~_~>)~2%"
	  (cons :export (all-exported-symbols))))


(defun print-header (outstream package)
  (print-defpackage outstream package)
  (format outstream "~&(in-package ~W)~2%" package)
  (format outstream "~&(declaim (optimize (speed 1) (compilation-speed 0) ~
                         (debug 2) (safety 2)))~2%")
  (format outstream "~&~{~W~%~}~%"
	  '((defmacro ior (&rest args)
	      (cond ((null args) 0)
		    ((null (rest args)) (first args))
		    (t
		     `(boole boole-ior ,(first args) (ior ,@(rest args))))))
	    (defparameter *c-interface-libraries*
	      '("-L/usr/X11R6/lib" "-lGL" "-lGLU" "-lXi" "-lXext"))
	    (defun load-ffi-object-file (name)
	      (ext:load-foreign name :libraries *c-interface-libraries*))
	    (eval-when (:load-toplevel :execute)
	      (load-ffi-object-file "/usr/lib/libGLU.so")))))

(defun print-dependencies (dependencies outstream)
  (dolist (dependency dependencies)
    (let ((ffi-object (type dependency)))
      (if ffi-object
	  (let ((alien (alien-form-for ffi-object)))
	    (when alien
	      (if (stringp alien)
		  (format outstream "~&~A~%" alien)
		  (format outstream "~&~W~%" alien))))
	  (warn "No FFI-Object named ~W." (type dependency))))))

(defmethod convert-file (package infile outfile)
  (clear-all-dependencies)
  (clear-all-ffi-tables)
  (parse-ffi-file infile)
  (process-all-ffi-objects)
  (let ((dependencies (sort-dependencies))
	(*print-case* :downcase))
    (with-open-file (outstream outfile
		     :direction :output :if-exists :supersede)
      (print-header outstream package)
      (print-dependencies dependencies outstream))))


(defparameter *c-interface-libraries*
  '("-L/usr/X11R6/lib" "-lGL" "-lGLU" "-lXi" "-lXext"))

(defun load-ffi-object-file (name)
  (ext:load-foreign name :libraries *c-interface-libraries*))

#||
(load-ffi-object-file "/usr/lib/libGLU.so")
(convert-file "GL"
	      "home:c-interface/ffi-files/glx.ffi"
	      "home:c-interface/output/glx.lisp")
(convert-file "GL"
	      "home:c-interface/ffi-files/glu.ffi"
	      "home:c-interface/output/glu.lisp")
(convert-file "GL"
	      "home:c-interface/ffi-files/X.ffi"
	      "home:c-interface/output/X.lisp")
(convert-file "GL"
	      "home:c-interface/ffi-files/Xlib.ffi"
	      "home:c-interface/output/Xlib.lisp")
||#
