;;; -*- Mode: LISP; Syntax: ANSI-Common-Lisp; Package: (C-INTERFACE); -*-

;;; C-Interface
;;; Copyright (C) 2002 Matthias Hlzl

;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License as
;;; published by the Free Software Foundation; either version 2 of the
;;; License, or (at your option) any later version.

;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.

;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
;;; 02111-1307, USA.

(in-package "C-INTERFACE")

(declaim (optimize (debug 3)))

;;; Dependencies.
;;; ============

(defvar *dependency-table* (make-hash-table :test #'equal))

 (defun clear-all-dependencies ()
  (clrhash *dependency-table*))

(defclass dependency ()
  ((type :accessor type :initarg :type)
   (depends-on :accessor depends-on
	       :initarg :depends-on :initform '())
   (is-dependency-for :accessor is-dependency-for
		      :initarg :is-dependency-for :initform '())
   ;; We copy the information in the Depends-On slot so that we can
   ;; destructively modify one copy when sorting the dependency table.
   (%depends-on :accessor %depends-on)))

;;; Find-Dependency.
;;; ===============

(defun find-dependency (type)
  (gethash type *dependency-table*))

(defun (setf find-dependency) (new-value type)
  (setf (gethash type *dependency-table*) new-value))

;;; Ensure-Dependency.
;;; =================

(defgeneric ensure-dependency (ffi-object))

(defmethod ensure-dependency ((object ffi-object))
  (let ((dependency (find-dependency object)))
    (or dependency
	(let ((new-dependency (make-instance 'dependency :type object)))
	  (setf (find-dependency object) new-dependency)
	  new-dependency))))

;;; Cyclic Dependencies.
;;; ===================

(defvar *cyclic-dependencies* '())

(defclass cyclic-dependency ()
  ((object-1 :accessor object-1 :initarg :object-1)
   (object-2 :accessor object-2 :initarg :object-2)))

(defun note-cyclic-dependency (object-1 object-2)
  (make-instance 'cyclic-dependency
		 :object-1 object-1
		 :object-2 object-2))


;;; Really-Note-Dependency.
;;; ======================

;;; This is the function that does all the dirty work once
;;; Note-Dependency has decided that finally wants to do something.

(defun really-note-dependency (type depends-on)
  (let ((dependency (ensure-dependency type))
	(dependency-2 (ensure-dependency depends-on)))
    (unless (eq type depends-on)
      (cond ((member depends-on (is-dependency-for dependency))
	     (note-cyclic-dependency type depends-on))
	    (t
	     (pushnew depends-on (depends-on dependency) :test #'equal)
	     (pushnew type (is-dependency-for dependency-2) :test #'equal))))))


;;; Note-Dependency.
;;; ===============

(defgeneric note-dependency (type depends-on))

;;; This is sometimes useful in the test-framework, when we convert
;;; forms without having set-up the whole dependency apparatus.  Maybe
;;; we should rewrite the tests and make this go away.
;;;
(defmethod note-dependency (object depends-on)
  (warn "Cannot add dependency from ~A to ~A."
	object depends-on))

(defmethod note-dependency
    ((object ffi-object) (depends-on null))
  (warn "Cannot add dependency from ~A to Nil."
	(name object)))

(defmethod note-dependency
    ((object ffi-object) (depends-on ffi-object))
  (really-note-dependency object depends-on))

;;; Dependencies on C-Type-References.
;;; ---------------------------------

(defmethod note-dependency
    ((object ffi-object) (ref primitive-c-type-reference))
  (declare (ignorable object ref))
  ;; We are not bothered by dependencies on primitive types.
  nil)

(defmethod note-dependency ((object ffi-object) (item field-item))
  (note-dependency object (item-type item)))

(defmethod note-dependency
    ((object ffi-object) (struct c-struct-type-reference))
  (note-dependency object (ffi-object-for-reference struct)))

(defmethod note-dependency
    ((object ffi-object) (union c-union-type-reference))
  (note-dependency object (ffi-object-for-reference union)))

(defmethod note-dependency
    ((object ffi-object) (enum c-enum-type-reference))
  (note-dependency object (ffi-object-for-reference enum)))

(defmethod note-dependency
    ((object ffi-object) (function c-function-type-reference))
  (note-dependency object (result-type function))
  (note-dependencies object (arglist function)))

(defmethod note-dependency
    ((object ffi-object) (ref derived-c-type-reference))
  (note-dependency object (base-type-reference ref)))

(defmethod note-dependency
    ((object ffi-object) (typedef c-typedef-reference))
  (note-dependency object (ffi-object-for-reference typedef)))

;;; Note-Dependencies.
;;; =================

(defun note-dependencies (type dependencies)
  (ensure-dependency type)
  (mapc (lambda (depends-on)
	  (note-dependency type depends-on))
	dependencies))

;;; Sort-Dependencies.
;;; =================

(defun find-initial-dependencies ()
  (loop for node being each hash-value of *dependency-table*
	when (null (depends-on node)) collect node))

(defun prepare-dependencies-for-copy ()
  (loop for node being each hash-value of *dependency-table*
	do (setf (%depends-on node) (depends-on node))))

(defvar *check-dependency-consistency* t)

(defun maybe-check-dependency-consistency (dependencies)
  (when *check-dependency-consistency*
    (assert (= (hash-table-count *dependency-table*)
	       (length dependencies))
	    ()
	    "Dependency-table has size ~A, dependencise ~A."
	    (hash-table-count *dependency-table*)
	    (length dependencies))))

(defun sort-dependencies ()
  (labels ((iterate (stack result)
	     (if (null stack)
		 result
		 (let* ((current-node (first stack))
			(new-stack (rest stack))
			(dependencies (is-dependency-for current-node)))
		   (mapc (lambda (node-name)
			   (let ((node (find-dependency node-name)))
			     (setf (%depends-on node)
				   (remove (type current-node) (%depends-on node)
					   :test #'equal))
			     (when (null (%depends-on node))
			       (push node new-stack))))
			 dependencies)
		   (iterate new-stack (cons current-node result))))))
    (prepare-dependencies-for-copy)
    (let ((dependencies (iterate (find-initial-dependencies) '())))
      (maybe-check-dependency-consistency dependencies)
      (nreverse dependencies))))

;;; Dependency Utilities.
;;; ====================

(defun show-dependency (node)
  (format t "~&~A depends-on: ~A :is-dependency-for: ~A~%"
	  (type node) (depends-on node) (is-dependency-for node)))

(defun show-dependencies (&optional (nr 100))
  (loop for node being each hash-value of *dependency-table*
	for n from 1 below nr
	do (show-dependency node)))

(defun find-missing-dependencies ()
  (let ((dependencies (sort-dependencies))
	(*check-dependency-consistency* nil))
    (loop for node being each hash-value of *dependency-table*
	  unless (member node dependencies)
	  collect node)))

(defun check-sorted-dependencies ()
  (let ((dependencies (sort-dependencies))
	(already-seen '()))
    (dolist (dependency dependencies)
      (push (type dependency) already-seen)
      (dolist (depends-on (depends-on dependency))
	(unless (member depends-on already-seen)
	  (format t "~&~A depends-on unknown ~A.~%"))))))