;;; -*- Mode: LISP; Syntax: ANSI-Common-Lisp; Package: (C-INTERFACE); -*-

;;; C-Interface
;;; Copyright (C) 2002 Matthias Hlzl

;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License as
;;; published by the Free Software Foundation; either version 2 of the
;;; License, or (at your option) any later version.

;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.

;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
;;; 02111-1307, USA.

(in-package "C-INTERFACE")

(declaim (optimize (debug 3)))

;;; Process-All-FFI-Objects.
;;; =======================

(defun process-all-ffi-objects ()
  (loop for ffi-objects being each hash-value of (get-ffi-object-table)
	do (mapc #'process-ffi-object ffi-objects)))

;;; Parse-And-Process-FFI-Form.
;;; ==========================

;;; This is a simple utility that is mainly used in the test suite and
;;; for interactive debugging.

(defun parse-and-process-ffi-form (form &key (type (type-for-form form)))
  "Parses Form into a FFI-Object and processes the result."
  (process-ffi-object (parse-ffi-form form :type type)))

;;; Process-FFI-Object.
;;; =================

(defgeneric process-ffi-object (ffi-object)
  (:documentation
   "Does all necessary processing for FFI-Object to faciliate the
generation of output by later stages."))

;;; The default method simply notes that we do not know how to process
;;; this kind of object.

(defmethod process-ffi-object ((ffi-object ffi-object))
  (setf (processing-state ffi-object) :not-implemented)
  ffi-object)

;;; A default around method makes sure that every FFI-Object apperas
;;; in our dependency table.

(defvar *print-processing-progress* t)

(defmethod process-ffi-object :around ((ffi-object ffi-object))
  (when *print-processing-progress*
    (format *standard-output* "~&Processing ~A." (name ffi-object))
    (force-output *standard-output*))
  (unless (eq (processing-state ffi-object) :done)
    (ensure-dependency ffi-object)
    (call-next-method)))

;;; Field-Mixin.
;;; -----------

(defun process-field (field)
  (destructuring-bind (name (%field type offset size)) field
    (assert (eq %field 'field))
    (make-instance 'field-item
		   :item-name (translate-c-name name)
		   :item-type (convert-ffi-type type)
		   :item-offset offset
		   :item-byte-size size)))

(defun process-fields (fields)
  (mapcar #'process-field fields))

(defmethod process-ffi-object :before ((ffi-object field-mixin))
  (let ((fields (process-fields (data ffi-object))))
    (setf (fields ffi-object) fields)
    (note-dependencies ffi-object fields)))

;;; FFI-Function.
;;; ------------

(defun process-ffi-function (function)
  (destructuring-bind (ffi-type args result) (data function)
    (assert (eq ffi-type 'function) ()
	    "Trying to process ~S as function." function)
    (let ((arglist (mapcar #'convert-ffi-type args))
	  (result-type (convert-ffi-type result)))
      (setf (arglist function) arglist
	    (result-type function) result-type)
      (note-dependencies function arglist)
      (note-dependency function result-type))))

(defmethod process-ffi-object ((function ffi-function))
  (when (eq (processing-state function) :unprocessed)
    (setf (processing-state function) :in-progress)
    (process-ffi-function function)
    (setf (processing-state function) :done))
  function)

;;; FFI-Variable.
;;; ------------

(defmethod process-ffi-object ((variable ffi-variable))
  (let ((base-type (convert-ffi-type (data variable))))
    (setf (base-type variable) base-type)
    (note-dependency variable base-type))
  (setf (processing-state variable) :done)
  variable)

;;; FFI-Typedef.
;;; -----------

(defmethod process-ffi-object ((typedef ffi-typedef))
  (let ((base-type (convert-ffi-type (data typedef))))
    (setf (base-type typedef) base-type)
    (note-dependency typedef base-type))
  (setf (processing-state typedef) :done)
  typedef)

;;; FFI-Struct.
;;; ----------

(defmethod process-ffi-object ((struct ffi-struct))
  (assert (slot-boundp struct 'fields))
  (setf (processing-state struct) :done)
  struct)

;;; FFI-Union.
;;; ---------

(defmethod process-ffi-object ((union ffi-union))
  (setf (processing-state union) :done)
  union)

;;; FFI-Enum.
;;; --------

(defun process-enum-item (item)
  (destructuring-bind (name value) item
    (make-instance 'enum-item
		   :item-name (translate-c-name name :package "KEYWORD")
		   :item-value value)))

(defun process-enum-items (items)
  (mapcar #'process-enum-item items))

(defmethod process-ffi-object ((enum ffi-enum))
  (setf (enum-items enum) (process-enum-items (data enum))
	(processing-state enum) :done)
  enum)

;;; FFI-Enum-Identifier.
;;; -------------------

(defmethod process-ffi-object ((identifier ffi-enum-identifier))
  (setf (processing-state identifier) :done)
  identifier)


;;; FFI-Macro.
;;; ---------

(defun macro-name-valid-p (name)
  (not (and (> (length name) 1)
            (string= (subseq name 0 2) "__"))))

;;; %%% Major Annoyance! 
;;; All this Compute-FOO-Expansion stuff should disappear and be
;;; replaced with a C-Expression parser.  But given the time
;;; constraints under which this program is written, I'll just
;;; punt and handle the important cases somehow.
;;;   -- tc 

(defun compute-cast-expansion (macro string)
  (let ((length (length string)))
    (when (> length 1)
      (cond
       ((find #\~ string :test #'char=)
        (let ((pos (position #\~ string :test #'char=)))
          (compute-simple-macro-expansion 
           macro (subseq string (1+ pos) (1- length)))))
       ;; Got somthing of the form "((unsigned int) 3)".
       ;; The position is 2 because the parser inserts a space.
       ((and (> length 2) (char= (char string 2) #\())
        (let* ((pos (position #\) string :test #'char=))
               (new-string (concatenate 'string "(" (subseq string (1+ pos)))))
          (compute-simple-macro-expansion macro new-string)))
       ;; Got something of the form "(int) 3".
       ((and (char= (char string 0) #\()
             (not (char= (char string (1- length)) #\))))
        (let* ((pos (position #\) string :test #'char= :from-end t))
               (new-string (subseq string (1+ pos))))
          (compute-simple-macro-expansion macro new-string)))
       (t nil)))))

(defvar *type-abbreviation-list*
  '(("int" (int ()))))

(defun type-abbreviation-p (type)
  (member type *type-abbreviation-list* :test #'string= :key #'car))

(defun converted-type (type)
  (let ((ffi-type
	 (cadar (member type *type-abbreviation-list* :test #'string= :key #'car))))
    (if ffi-type
	(convert-ffi-type ffi-type)
	nil)))

(defmethod maybe-warn-for-typedef-redefinition
    ((old-object ffi-object) (macro ffi-macro))
  (warn "Changing macro ~A to typedef would result in conflict with ~A."
	(name macro) old-object))

(defmethod maybe-warn-for-typedef-redefinition
    ((old-object ffi-typedef) (macro ffi-macro))
  (warn "Not changing macro ~A to typedef because a typedef with the same ~
         name already exists."
	(name macro)))

(defun change-macro-to-type (macro data)
  (let ((converted-type (converted-type data)))
    (setf (find-ffi-macro (lisp-name macro)) nil)
    (setf (find-ffi-object (lisp-name macro))
	  (remove macro (find-ffi-object (lisp-name macro))))
    ;;; Strip leading and trailing #\+ off the lisp-name.
    (setf (lisp-name macro) (translate-c-name (name macro)))
    (cond ((find-ffi-object (lisp-name macro))
	   (maybe-warn-for-typedef-redefinition
	    (first (find-ffi-object (lisp-name macro)))
	    macro)
	   (setf (processing-state macro) :deleted))
	  (t
	   (setf (find-ffi-typedef (lisp-name macro))
		 (list* macro (find-ffi-typedef (lisp-name macro))))
	   (setf (find-ffi-object (lisp-name macro))
		 (list* macro (find-ffi-object (lisp-name macro))))
	   (change-class macro 'ffi-typedef)
	   (setf (base-type macro) converted-type)
	   ;; This becomes necessary only when we refine the type
	   ;; changing process to recognize macros that expand into
	   ;; typedefs or stuff like that, but it does not hurt us at
	   ;; the moment, so we'll leave it in.
	   (note-dependency macro converted-type)
	   (setf (processing-state macro) :done)))))

;;; Compute-Simple-Macro-Expansion returns
;;;   * Nil if it doesn't know how to process the macro
;;;   * A value, if if can compute a macro expansion
;;;   * The magic value :Done, if we found out that somebody used a
;;;     macro as typedef or in some other nasty way and we therefore
;;;     changed the FFI-Macro class to something else.

(defun compute-simple-macro-expansion (macro data)
  (assert (stringp data) () "Macro data ~S is not a string." data)
  (cond ((string= data "") t)
	((char= (char data 0) #\") data)
	((compute-ior-expression macro data))
	((compute-cast-expansion macro data))
	((type-abbreviation-p data)
	 (change-macro-to-type macro data)
	 :done)
	((find #\Space data :test #'char=)
	 (compute-list-expansion macro (split-string '(#\Space) data)))
	((try-to-parse-as-number data))
        ;;; %%% Don't know what to do with strings.
	(t data)))

(defun compute-ior-expression (macro string)
  (let ((length (length string)))
    (when (and (> length 2) (find #\| string :test #'char=))
      (let* ((string (string-trim '(#\( #\)) string))
             (strings (mapcar (lambda (str)
                                (string-trim '(#\Space) str))
                              (split-string '(#\|) string))))
        `(ior ,@(mapcar (lambda (string)
			  (or (try-to-parse-as-number string)
			      (let ((c-name (translate-c-name string
							      :type 'ffi-macro)))
				(mapc (lambda (object)
					(note-dependency macro object))
				      (find-ffi-object c-name))
				c-name)))
			strings))))))

(defun compute-list-expansion (macro list)
  (cond ((null list) nil)
        ;; Had leading or trailing spaces...
        ((singleton-list-p list)
         (let ((first (first list)))
           (assert (stringp first))
           (compute-simple-macro-expansion macro first)))
        (t
         (case (length list)
           (5 (when (and (string= "(" (first list))
                         (string= "<<" (third list))
                         (string= ")" (fifth list)))
                (let ((base (try-to-parse-as-number (second list)))
                      (shift-amount (try-to-parse-as-number (fourth list))))
                  (when (and base shift-amount)
                    (ash base shift-amount)))))
           (3 (when (and (string= "(" (first list))
                         (string= ")" (third list)))
                (compute-simple-macro-expansion macro (second list))))
           (2 (when (string= "-" (first list))
                (let ((num (try-to-parse-as-number (second list))))
                  (when num
                    (- num)))))
           (default nil)))))

;;; A simple ffi macro is a macro that does not take parameters.
;;; This is the only kind we know how to handle at the moment.

(defun process-simple-ffi-macro (macro)
  (if (macro-name-valid-p (name macro))
      (let ((exp (compute-simple-macro-expansion macro (data macro))))
	(cond ((eq exp :done)
	       ;; We changed the macro to something else.  Do nothing.
	       nil)
	      (exp
	       (setf (expansion macro) exp
		     (processing-state macro) :done))
	      (t
	       (setf (expansion macro) nil
		     (processing-state macro) :failed))))
      (setf (expansion macro) nil
	    (processing-state macro) :not-implemented)))

(defmethod process-ffi-object ((macro ffi-macro))
  (when (eq (processing-state macro) :unprocessed)
    (setf (processing-state macro) :in-progress)
    (if (find #\Space (name macro) :test #'char=)
	;; If Name contains a space we have a complex macro.
	;; Give up. (Might be Process-Complex-FFI-Macro some
	;; day.)
	(call-next-method)
        (process-simple-ffi-macro macro)))
  macro)
