;;; -*- Mode: LISP; Syntax: ANSI-Common-Lisp; Package: (C-INTERFACE); -*-

;;; C-Interface
;;; Copyright (C) 2002 Matthias Hlzl

;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License as
;;; published by the Free Software Foundation; either version 2 of the
;;; License, or (at your option) any later version.

;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.

;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
;;; 02111-1307, USA.

(in-package "C-INTERFACE")

(define-fixture classes-fixture ()
  ((simple-macro :accessor simple-macro)))

(defmethod set-up ((fixture classes-fixture))
  (clear-all-ffi-tables)
  (setf (simple-macro fixture)
	(make-instance 'ffi-macro :name "SIMPLE_MACRO" :lisp-name '+simple-macro+)))

(define-test-case classes-test-case
                  :parent 'c-interface
                  :fixture-class 'classes-fixture)

(set-current-test-case classes-test-case)

(deftest (macro-name-valid-p)
  (expect-true (macro-name-valid-p ""))
  (expect-true (macro-name-valid-p "_"))
  (expect-true (macro-name-valid-p "_foo"))
  (expect-false (macro-name-valid-p "__"))
  (expect-false (macro-name-valid-p "__foo"))
  (expect-false (macro-name-valid-p "___foo")))

(deftest (type-abbreviation-p)
  (expect-false (type-abbreviation-p "integer"))
  (expect-true (type-abbreviation-p "int")))

(deftest (converted-type)
  (expect-type (converted-type "int") 'c-int)
  (expect (converted-type "foo") nil))

(deftest (compute-list-expansion)
  (expect (compute-list-expansion nil ()) nil)
  (expect (compute-list-expansion nil '("123")) 123)
  (expect (compute-list-expansion nil '("(" "123" ")")) 123)
  (expect (compute-list-expansion nil '("-" "123")) -123)
  (expect (compute-list-expansion nil '("-" "foo")) nil)
  (expect (compute-list-expansion nil '("(" "1" "<<" "0" ")")) 1)
  (expect (compute-list-expansion nil '("(" "1" "<<" "1" ")")) 2)
  (expect (compute-list-expansion nil '("(" "1" "<<" "2" ")")) 4)
  (expect (compute-list-expansion nil '("(" "1" "<<" "3" ")")) 8)
  (expect (compute-list-expansion nil '("(" "3" "<<" "1" ")")) 6)
  (expect (compute-list-expansion nil '("(" "3" "<<" "3" ")")) 24))

(deftest (compute-simple-macro-expansion)
  (expect-condition error (compute-simple-macro-expansion nil ()))
  (expect (compute-simple-macro-expansion nil "") t)
  (expect (compute-simple-macro-expansion nil "123") 123)
  (expect (compute-simple-macro-expansion nil "0xfd") 253)
  (expect (compute-simple-macro-expansion nil "( 123 )") 123)
  (expect (compute-simple-macro-expansion nil "( 0x0004 )") 4)
  (expect (compute-simple-macro-expansion nil "- 123") -123)
  (expect (compute-simple-macro-expansion nil "( 1 << 0 )") 1)
  (expect (compute-simple-macro-expansion nil "( 1 << 1 )") 2)
  (expect (compute-simple-macro-expansion nil "( 1 << 2 )") 4)
  (expect (compute-simple-macro-expansion nil "( 1L << 4 )") 16))

(deftest (compute-ior-expression)
  (expect (handler-bind ((warning #'muffle-warning))
	    (compute-ior-expression simple-macro "1|2"))
	  '(ior 1 2) :test #'equal)
  (expect (handler-bind ((warning #'muffle-warning))
	    (compute-ior-expression simple-macro "1 | 2"))
	  '(ior 1 2) :test #'equal)
  (expect (handler-bind ((warning #'muffle-warning))
	    (compute-ior-expression simple-macro "foo|bar"))
	  '(ior +foo+ +bar+) :test #'equal)
  (expect (handler-bind ((warning #'muffle-warning))
	    (compute-ior-expression simple-macro "foo|bar|baz|fooBar"))
          '(ior +foo+ +bar+ +baz+ +foo-bar+) :test #'equal))

;;; Some utilities to test the parser.

(defun expect-ffi-stuff (expected-name expected-lisp-name expected-data
                         expected-source)
  (let ((object (first (find-ffi-object expected-lisp-name))))
    (if object
	(with-slots (name lisp-name data source-location) object
	  (expect source-location expected-source :test #'equal)
	  (expect name expected-name :test #'equal)
	  (expect lisp-name expected-lisp-name)
	  (expect data expected-data :test #'equal))
	(error "No FFI object named ~A." expected-lisp-name))))

(defun expect-ffi-typedef (expected-name expected-lisp-name expected-data
			   expected-source expected-alien-type)
  (expect-ffi-stuff expected-name expected-lisp-name expected-data expected-source)
  (let ((type (first (find-ffi-typedef expected-lisp-name))))
    (expect-type type 'ffi-typedef)
    (expect-type (base-type type) expected-alien-type)))

(defun expect-ffi-macro (expected-name expected-lisp-name expected-data
                         expected-expansion expected-source)
  (expect-ffi-stuff expected-name expected-lisp-name expected-data expected-source)
  (let ((macro (first (find-ffi-macro expected-lisp-name))))
    (expect-type macro 'ffi-macro)
    (with-slots (expansion) macro
      (expect expansion expected-expansion :test #'equal))))

(deftest (parse-ffi-macro/1)
  (clear-ffi-macro-table)
  (expect-false (find-ffi-macro "GXor"))
  (parse-and-process-ffi-form '(macro ("/usr/include/X11/X.h" 488) "GXor" "0x7")
			      :type 'ffi-macro)
  (expect-ffi-macro "GXor" '+g-xor+ "0x7" 7 
                    '("/usr/include/X11/X.h" 488)))

(deftest (parse-ffi-macro/2)
  (clear-ffi-macro-table)
  (parse-and-process-ffi-form '(macro ("/usr/include/X11/X.h" 519)
				"FillSolid" "0"))
  (expect-ffi-macro "FillSolid" '+fill-solid+ "0" 0 
                    '("/usr/include/X11/X.h" 519)))

(deftest (parse-ffi-macro/3)
  (clear-ffi-macro-table)
  (parse-and-process-ffi-form '(macro ("/usr/include/X11/X.h" 156)
				"EnterWindowMask" "( 1L << 4 )"))
  (parse-and-process-ffi-form '(macro ("/usr/include/X11/X.h" 619)
				"DoGreen" "( 1 << 1 )"))
  (expect-ffi-macro "EnterWindowMask" '+enter-window-mask+ "( 1L << 4 )" 16
                    '("/usr/include/X11/X.h" 156))
  (expect-ffi-macro "DoGreen" '+do-green+ "( 1 << 1 )" 2
                    '("/usr/include/X11/X.h" 619)))

(deftest (parse-ffi-macro/4)
  (clear-ffi-macro-table)  
  (parse-and-process-ffi-form '(macro ("/usr/include/X11/Xlib.h" 1224)
				"XBufferOverflow" "- 1"))
  (parse-and-process-ffi-form '(macro ("/usr/include/X11/Xlib.h" 1309)
				"XIMStringConversionChar" "( 0x0004 )"))
  (expect-ffi-macro "XBufferOverflow" '+x-buffer-overflow+ "- 1" -1
                    '("/usr/include/X11/Xlib.h" 1224))
  (expect-ffi-macro "XIMStringConversionChar" '+xim-string-conversion-char+
                    "( 0x0004 )" 4
                    '("/usr/include/X11/Xlib.h" 1309)))

(deftest (parse-ffi-macro/5)
  (clear-ffi-macro-table)
  (parse-and-process-ffi-form '(macro ("/usr/include/X11/Xlib.h" 1219)
				"XNHotKey" "\"hotKey\""))
  (expect-ffi-macro "XNHotKey" '+xn-hot-key+ "\"hotKey\"" "\"hotKey\""
                    '("/usr/include/X11/Xlib.h" 1219)))

(deftest (parse-ffi-macro/6)
  (clear-ffi-macro-table)
  (parse-and-process-ffi-form '(macro ("/usr/include/X11/Xlib.h" 92)
				"Bool" "int"))
  (expect-ffi-typedef "Bool" 'bool "int"
		      '("/usr/include/X11/Xlib.h" 92)
		      'c-int))
  
(deftest (parse-ffi-macro/7)
  (clear-ffi-macro-table)
  (parse-and-process-ffi-form '(macro ("/usr/include/X11/Xlib.h" 109)
				"AllPlanes" "( ( unsigned long ) ~ 0L )"))
  (expect-ffi-macro "AllPlanes" '+all-planes+ "( ( unsigned long ) ~ 0L )" 0
                    '("/usr/include/X11/Xlib.h" 109)))

(deftest (parse-ffi-macro/8)
  (clear-ffi-macro-table)
  (parse-and-process-ffi-form '(macro ("/usr/include/linux//stddef.h" 8)
				"NULL" "( ( void * ) 0 )"))
  (expect-ffi-macro "NULL" '+null+ "( ( void * ) 0 )" 0
                    '("/usr/include/linux//stddef.h" 8)))

(deftest (parse-ffi-macro/9)
  (clear-ffi-macro-table)
  (parse-and-process-ffi-form '(macro ("/usr/include//X11/X.h" 339) 
				"RevertToPointerRoot" "( int ) PointerRoot"))
  (expect-ffi-macro "RevertToPointerRoot" '+revert-to-pointer-root+
                    "( int ) PointerRoot" "PointerRoot"
                    '("/usr/include//X11/X.h" 339)))

(deftest (sort-dependencies/1)
  (clear-all-ffi-tables)
  (clear-all-dependencies)
  (expect (hash-table-count *dependency-table*) 0))

#||
(deftest (sort-dependencies/2)
  (clear-all-ffi-tables)
  (clear-all-dependencies)
  ;;; This might fail if you have different paths...
  (parse-ffi-file "home:c-interface/ffi-files/glx.ffi")
  (process-all-ffi-objects)
  (expect-true (> (hash-table-count *dependency-table*) 0))
  (expect-true (= (length (sort-dependencies))
		  (hash-table-count *dependency-table*))))
||#
#||
(show-results (run-tests 'c-interface))
||#
