;;; -*- Mode: LISP; Syntax: ANSI-Common-Lisp; Package: (C-INTERFACE); -*-

;;; C-Interface
;;; Copyright (C) 2002 Matthias Hlzl

;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License as
;;; published by the Free Software Foundation; either version 2 of the
;;; License, or (at your option) any later version.

;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.

;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
;;; 02111-1307, USA.

(in-package "C-INTERFACE")

;;; Stuff that should mostly be in another package.

(defmacro ior (&rest args)
  (cond ((null args) 0)
        ((null (rest args)) (first args))
        (t
         `(boole boole-ior ,(first args) (ior ,@(rest args))))))

(defun listify (stuff)
  (if (listp stuff)
      stuff
    (list stuff)))

(defun list-of (n thing)
  (if (zerop n)
      '()
      (cons thing (list-of (1- n) thing))))

(defun singleton-list-p (stuff)
  (and (consp stuff) (null (rest stuff))))

(defun pair-equal (pair-1 pair-2)
  (or (and (equal (car pair-1) (car pair-2))
	   (equal (cdr pair-1) (cdr pair-2)))
      (and (equal (car pair-1) (cdr pair-2))
	   (equal (cdr pair-1) (car pair-2)))))

(defun split-string (character-replacement-bag string)
  "Splits String at every occurence of an element of
   Character-Replacement-Bag.  If an element of the
   Character-Replacement-Bag is a cons, its :cdr is inserted into the
   result."
  (let ((replacement-bag (mapcar (lambda (arg)
                                   (if (consp arg) arg (cons arg nil)))
                                 character-replacement-bag))
        (start 0)
        (length (length string))
	(result '()))
    (loop for index from 0 to length do
	  (cond ((= length index)
		 ;; We don't collect an empty string when we are done,
		 ;; unless the original string was empty to begin with.
		 (unless (and (> start 0) (= start index)) 
		   (let ((substring (subseq string start index)))
		     (push substring result))))
		(t
		 (let ((char (char string index)))
		   (let ((split (assoc char replacement-bag)))
		     (when split
		       (unless (= index start)
			 (let ((substring (subseq string start index)))
			   (push substring result)))
		       (setf start (1+ index))
		       (when (cdr split)
			 (push (cdr split) result))))))))
    (nreverse result)))

;;; We keep a buffer for Unstudlify so that we don't allocate a new
;;; string for every conversion.

(defvar *unstudlify-buffer* (make-array 1024 :element-type 'character :fill-pointer 0))

;;; Unstudlify is somewhat complicated, because we want to get stuff
;;; like FooBARBaz right (it should convert to foo-bar-baz).  This is
;;; essentially a state machine driven by the change between upper-
;;; and lower-case.  When a name contains an underscore we just
;;; translate underscores to dashes.

(defun translate-underscores (string)
  (substitute #\- #\_ (string-upcase string) :test #'equal))

(defun unstudlify (string &key (prefix "") (postfix ""))
  (multiple-value-bind (new-string new-prefix new-postfix)
      (preprocess-string string prefix postfix)
    (really-unstudlify new-string new-prefix new-postfix)))

(defparameter *known-prefixes*
  '(("___" "%%%")
    ("__" "%%")
    ("_" "%")
    ("GLU_" "GLU-")
    ("GLU" "GLU-")
    ("GLX_" "GLX-")
    ("GLX" "GLX-")
    ("glX" "GLX-")
    ("GL_" "GL-")
    ("GL" "GL-")
    ("XUTF8" "XUTF8-")
    ("XIM" "XIM-")
    ("XNHot" "XN-HOT-")
    ))

(defun known-prefix (string)
  (let ((length-string (length string)))
    (dolist (pair *known-prefixes*)
      (let* ((prefix (first pair))
	     (length-prefix (length prefix)))
	(when (and (<= length-prefix length-string)
		   (string= prefix string :end2 length-prefix))
	  (return-from known-prefix
	    (values (second pair) length-prefix))))))
  (values nil nil))
    
(defun preprocess-string (string prefix postfix)
  (multiple-value-bind (known-prefix index)
      (known-prefix string)
    (cond (known-prefix
	   (values (subseq string index)
		   (concatenate 'string prefix known-prefix)
		   postfix))
	  ((and (not (string= "" string))
		(every #'digit-char-p string))
	   (values string
		   "ANONYMOUS-"
		   postfix))
	  (t
	   (values string
		   prefix
		   postfix)))))

(defun really-unstudlify (string prefix postfix)
  (cond ((equal string "")
         (concatenate 'string prefix postfix))
        ((= (length string) 1)
         (concatenate 'string prefix (string-upcase string) postfix))
        (t
         (if (find #\_ string :test #'equal)
             (concatenate 'string prefix (translate-underscores string) postfix)
           (let ((buffer *unstudlify-buffer*))
             (setf (fill-pointer buffer) 0)
             (loop for char across prefix
		   do (vector-push-extend char buffer))
             (unstudlify-state-machine string)
             (loop for char across postfix
		   do (vector-push-extend char buffer))
             (copy-seq buffer))))))

(defun unstudlify-state-machine (string)
  (let ((buffer *unstudlify-buffer*)
        ;; Push-Dash-P is true if we still need to insert a dash
        ;; because we have read an upper-case char but not yet
        ;; output a dash.
        (push-dash-p nil)
        ;; Push-Char is either Nil or an uppercase char whose
        ;; output we have deferred because we don't know whether
        ;; to precede it with a dash or not.
        (push-char nil))
    (vector-push-extend (char-upcase (char string 0)) buffer)
    (loop for char across (subseq string 1)
	  and prev = (char string 0) then char
	  do (cond ((upper-case-p char)
		    (cond ((not (upper-case-p prev))
			   ;; First upper-case char, we have to output a dash in
			   ;; every case.
			   (vector-push-extend #\- buffer)
			   (setf push-char char)
			   (setf push-dash-p nil))
			  (push-char
			   ;; We have an upper-case char and are already pushing chars.
			   (vector-push-extend (char-upcase push-char) buffer)
			   (setf push-char char)
			   (setf push-dash-p t))
			  (t
			   ;; This case only happens for strings where the first 
			   ;; two characters are both upper case (I think).
			   (setf push-char char)
			   (setf push-dash-p t))))
		   ((upper-case-p prev)
		    (when (and push-dash-p prev)
		      (vector-push-extend #\- buffer)
		      (setf push-dash-p nil)
		      (assert (equal prev push-char)))
		    (when push-char
		      (assert (equal prev push-char))
		      (vector-push-extend (char-upcase push-char) buffer)
		      (setf push-char nil))
		    (vector-push-extend (char-upcase char) buffer))
		   (t
		    (assert (and (not push-dash-p) (not push-char)))
		    (vector-push-extend (char-upcase char) buffer))))
    (when push-char
      (vector-push-extend (char-upcase push-char) buffer))))

(defvar *c-name-translations* (make-hash-table :test #'equal))

(defun clear-c-name-translations ()
  (clrhash *c-name-translations*))

(defun add-c-translation (c-name lisp-name)
  (setf (gethash c-name *c-name-translations*) lisp-name))

(defun find-c-translation (c-name)
  (gethash c-name *c-name-translations* nil))

(defmacro define-c-translation (c-name lisp-name)
  `(add-c-translation ,c-name ,lisp-name))

(defun translate-c-name (name &key (package *package*) type (prefix "") (postfix ""))
  (if (eq type 'ffi-macro)
      (translate-c-name name :package package :prefix "+" :postfix "+")
      (or (find-c-translation name)
	  (intern (unstudlify name :prefix prefix :postfix postfix)
		  package))))

(defun valid-number-or-nil (number pos string)
  (when number
    (let ((length (length string)))
      (when (or (= pos length)
                (and (= pos (1- length))
                     (char-equal (char string (1- length)) #\L)))
        number))))

(defun try-to-parse-as-decimal-number (string)
  (multiple-value-bind (result pos) 
      (parse-integer string :junk-allowed t)
    (valid-number-or-nil result pos string)))

(defun try-to-parse-as-hexadecimal-number (string)
  (multiple-value-bind (result pos)
      (parse-integer string :junk-allowed t :radix 16 :start 2)
    (valid-number-or-nil result pos string)))

(defun try-to-parse-as-number (string)
  (unless (equal string "")
    (if (and (> (length string) 1)
             (eql (char string 0) #\0)
             (char-equal (char string 1) #\x))
        (try-to-parse-as-hexadecimal-number string)
      (try-to-parse-as-decimal-number string))))

(defgeneric tokenize (thing))

(defmethod tokenize ((string string))
  (or (try-to-parse-as-number string) string))

(defmethod tokenize ((list list))
  (cond ((null list)
         ())
        ((singleton-list-p list)
         (let ((token (first list)))
           (if (stringp token)
               (or (try-to-parse-as-number token)
                   token)
             token)))
        (t
         (error "Cannot tokenize lists."))))

(defun parse-macro (macro-form)
  (destructuring-bind (macro source-location name body)
      macro-form
    (declare (ignore source-location))
    (assert (eq macro 'macro) ()
      "Trying to parse ~A as macro form." macro-form)
    (let ((parsed-body (tokenize body)))
      (if (numberp parsed-body)
          `(defconstant ,(translate-c-name name :prefix "+" :postfix "+")
             ,parsed-body)))))

;;; Structure conversion.
;;;; ===================

(defvar *dont-convert-structs*
  '(funcs)
  "A list of structure names that should be replaced with int[] definitions.")

(defun ignored-struct-p (struct)
  (member struct *dont-convert-structs*))


(defmacro define-hashed-class (name supers slots &rest args)
  (let* ((class-hash (gensym))
         (sname (symbol-name name))
         (spackage (symbol-package name))
         (finder-name (intern (concatenate 'string "FIND-" sname) spackage))
         (clearer-name (intern (concatenate 'string "CLEAR-" sname "-TABLE")
                               spackage))
         (hash-getter (intern (concatenate 'string "GET-" sname "-TABLE")
                              spackage)))
    `(progn
      (defclass ,name ,supers
	,slots
	,@args)
      (defvar ,class-hash (make-hash-table :test #'equal))
      (defun ,clearer-name ()
	(declare (special ,class-hash))
	(clrhash ,class-hash))
      ;; This is mostly useful for debugging purposes.
      (defun ,hash-getter ()
	(declare (special ,class-hash))
	,class-hash)
      ;; The following two definitions are generic so that it is easy to
      ;; define around methods etc.
      (defgeneric ,finder-name (id))
      (defmethod ,finder-name (id)
	(declare (special ,class-hash))
	(gethash id ,class-hash nil))
      (defgeneric (setf ,finder-name) (obj id))
      (defmethod (setf ,finder-name) (obj id)
	(declare (special ,class-hash))
	(setf (gethash id ,class-hash) obj))
      (defmethod initialize-instance :after ((this ,name) &key (id nil))
	(when id
	  (setf (,finder-name id)
		(list* this (,finder-name id)))))
      ',name)))

;;; SAP Utilities.
;;; =============

(defun sap-accessor-for-byte-size (byte-size)
  (case byte-size
    (1 'system:sap-ref-8)
    (2 'system:sap-ref-16)
    (4 'system:sap-ref-32)
    (8 'system:sap-ref-64)
    (t nil)))

;;; Inspect is broken in current CMUCL builds.

(defun object-slot-names (object)
  (mapcar #'pcl::slot-definition-name  (pcl:class-slots (pcl::class-of object))))

(defun inspect-object (object)
  (format t "~&Object ~A~%" object)
  (mapc (lambda (slot)
	  (when (slot-boundp object slot)
	    (format t "  ~A: ~W~%" slot (slot-value object slot))))
	(object-slot-names object)))
